import java.lang.*;
import java.util.*;
import java.io.*;
import org.w3c.dom.*;

/*
   This software may be freely used for academic purpose only. Please
   contact us for profit use. However, when distributed, the original
   source must be clearly stated, and, when the source code is
   distributed, the copiright notice must be retained and any
   alterations in the code must be clearly marked. No warranty is
   given regarding the quality of this software.
*/

/** A wrapper class for ArrayList */
class PatternList extends ArrayList {
    PatternList() {
	super();//Note: Not neccesary to be specified; default;
    }

    PatternList(PatternList list) {
	super((ArrayList)list);
    }
}

/** A wrapper class for ArrayList */
class NodeList extends ArrayList {
    NodeList() {
	super(); //Note: Not neccesary to be specified; default;
    }

    NodeList(NodeList list) {
	super((ArrayList)list);
    }
}

public class Pattern{
    //Constants
    static boolean DEBUG = false;
    static boolean PRINT_STYLE_FOLD = true;
    static int gId = 0;

    //Fields
    Snode rmleaf_;		//A labeled tree specified by the rightmost leaf
    int   rmblen_;  	//The length of the rightmost branch
    int   size_;
    int   id_; 
    NodeList rmolist_; 		//The list of the rightmost occurrences
    NodeList neg_rmolist_; 	//For optimized mining
    int rootocc_; 		//The root occurrence of the pattern 
    int neg_rootocc_;        	//For optimized mining
    boolean is_closed; 		//For closed pattern mining

    Pattern(){
	//Create an empty tree
	initializeAll();
	rmolist_ = new NodeList();
	neg_rmolist_ = new NodeList();
	size_ = 0;
	id_   = gId++; 
    }

    Pattern(Snode node){
	//Create a single node pattern with a given label
	initializeAll();
	rmleaf_  = node; 
	rmolist_ = new NodeList();
	neg_rmolist_ = new NodeList();
	size_ = 1;
	id_   = gId++; 
    }

    //Only for compatibility
    Pattern(String nodename){
	initializeAll();
	rmleaf_  = new Snode(nodename);
	rmolist_ = new NodeList();
	neg_rmolist_ = new NodeList();
	size_ = 1;
	id_   = gId++; 
    }

    Pattern(Pattern pat) {
	rmleaf_       = copyTreeAtRmleaf(pat.rmleaf_);
	rmblen_       = pat.rmblen_;
	rmolist_      = new NodeList(pat.rmolist_);
	neg_rmolist_  = new NodeList(pat.neg_rmolist_);
	rootocc_      = pat.rootocc_;
	neg_rootocc_  = pat.neg_rootocc_;
	id_   = gId++; 
    }

    void initializeAll() {
	rmleaf_       = null;
	rmblen_ = 0;
	size_          = 0;
	rmolist_      = null;
	neg_rmolist_  = null;
	rootocc_      = 0;
	neg_rootocc_  = 0;
	is_closed     = true; //for closed patterns
    }

    void copySkeltonOnly(Pattern pat) {
	if (pat == null)
	    Util.errorExit("copySkeltonOnly: pat==null!");
	rmleaf_       = copyTreeAtRmleaf(pat.rmleaf_);
	rmblen_       = pat.rmblen_;
	size_         = pat.size_;
	rmolist_      = rmolist_;     //dammy: Do not change old one
	neg_rmolist_  = neg_rmolist_; //dammy: Do not change old one
	rootocc_      = 0; 
	neg_rootocc_  = 0; 
    }



    /************************************************************
     * Accessors
     ************************************************************/

    //public Snode getPatRoot(){
    public Snode getRmLeaf(){ return rmleaf_; }

    public int getRmbLength(){ return rmblen_; }

    public int getSize(){ return size_; }

    public int getId(){ return id_; }

    public NodeList getRmoList(){ return rmolist_; }

    public NodeList getNegNodeList(){ return neg_rmolist_; }

    public int getRootOcc(){ return rootocc_; }
    public void setRootOcc(int i){ rootocc_ = i; }

    public int getNegRootOcc(){ return neg_rootocc_; }
    public void setNegRootOcc(int i){ neg_rootocc_ = i; }

    public int getRmOcc(){ return rmolist_.size(); }
    public int getNegRmOcc(){ return neg_rmolist_.size(); }

    //For closed pattern mining
    public boolean getClosedFlag(){ return is_closed; }
    public void setClosedFlag(boolean value){ is_closed = value; }

    /************************************************************
     * 
     ************************************************************/

    public Snode getRootNode(){
	return fetchRootNode(rmleaf_);
    }

    /** Return the root node above the given node */
    static public Snode fetchRootNode(Snode node){
        while (node.parent != null)
            node = node.parent;
        return node;
    }

    /** Return the rightmost leaf (RML) of the given pattern with the root.
     *  where RML is the node of the last node in the preorder
     *  traversal of the pattern */
    static public Snode fetchRML(Snode root){
        Snode node = root; 
        while (node.firstchild != null){
            node = node.firstchild;
            while (node.nextsibling != null)
                node = node.nextsibling;
        }
        return node;
    }

    /************************************************************
     * Rightmost Expansion: 
     ************************************************************/

    private Snode getLastSibling(Snode node) {
	while(node.nextsibling != null)
	    node = node.nextsibling;
	return node;
    }

    /** Attaching a new node to the parent pattern as its rightmost leaf.
     *  This is the only operation to modify a pattern.
     */
    private Snode appendChild(Snode parent, Snode child){
	if (parent == null)
	    Util.errorExit("Pattern.appendChild: parent==null!");
	if (child == null)
	    Util.errorExit("Pattern.appendChild: child==null!");
        if (parent.firstchild == null)
            parent.firstchild = child;
        else {
	    Snode last = getLastSibling(parent.firstchild);
            last.nextsibling = child;
	}
	child.parent = parent;
	if (DEBUG) System.out.print("appendChild: "+size_);
	size_++;
	if (DEBUG) System.out.println(" => "+size_);
        return child;
    }

    //Node version: 
    public void expand(int depth, Snode newNode){
	if (newNode == null)
	    Util.errorExit("Pattern.expand: null node: "+newNode);
	if (depth < 0 || depth > rmblen_)
	    Util.errorExit("Pattern.expand: pair <"+
			   depth+", "+newNode.getNodeName()+">"+
			   " must be within the range ["+0+".."+rmblen_+"]");
	if (DEBUG)
	    System.out.println("Pattern.expand: expanding with "+
			   "<"+depth+", \""+newNode.getNodeName()+"\">");

	if (newNode.getNodeValue() != null)
	    newNode.setNodeValue(null);
	Snode node = rmleaf_;
	if (node == null) Util.errorExit("Pattern.expand[1]: node==null!");

	//Attaching the new node to the RM-leaf
        if (depth == 0){
            appendChild(node, newNode);
        }
	else {
	    //node = node.parent;
	    //depth--;
	    int d = depth; 
	    while (d > 0){
		if (node == null)
		    Util.errorExit("Pattern.expand[2]: node==null (d="+d+")");
		node = node.parent;
		d--;
	    }
	    if (node == null)
		Util.errorExit("Pattern.expand[3]: node==null (d="+d+")");
	    appendChild(node, newNode);
	}
	
	//Updating RM-branch length
	if (DEBUG) System.out.println("@debug Pattern.expand: update rmblen"+
			   "rmblen="+rmblen_+
			   ", depth="+depth+
			   ", (rmblen - depth + 1)="+(rmblen_ - depth + 1));
	rmblen_ = rmblen_ - depth + 1;
	rmleaf_ = newNode;
    }

    public void addRmOcc(Node node){
	rmolist_.add(node);
    }

    public void addNegRmOcc(Node node){
	neg_rmolist_.add(node);
    }

    /************************************************************
     *
     ************************************************************/

    Snode fetchPthParent(int depth) {
	Snode node = rmleaf_;
	int d = depth; 
	while (d > 0){
	    if (node == null)
		Util.errorExit("Pattern.fetchPthParent: node==null (d="+d+")");
	    node = node.parent;
	    d--;
	}
	if (node == null)
	    Util.errorExit("Pattern.fetchPthParent: node==null (d="+d+")");
	return node;
    }

    /** Create a copy of a given tree and then returns its Rmleaf */
    static public Snode copyTreeAtRmleaf(Snode rmleaf) {
	if (rmleaf == null)
	    Util.errorExit("copyTreeAtRmleaf: rmleaf==null!");
	Snode oldroot = fetchRootNode(rmleaf);
	Snode newroot = copyTreeAtRoot(oldroot);
	return fetchRML(newroot);
    }

    /* Warning: The sound implemntation of the procedure copyTreeAtRoot()
     * is critical to the correctness
     */
    
    /** Create the copy of the subtree rooted at parent and
     *  return its root
     */
    static public Snode copyTreeAtRoot(Snode parent) {
	if (parent == null)
	    Util.errorExit("copyTreeAtRoot: parent==null!");

	Snode child, prev; 
	Snode parentNew, childNew, prevNew; 

	parentNew = parent.copy();
	parentNew.firstchild = null;
	parentNew.nextsibling = null;
	parentNew.parent = null;

	child  = parent.firstchild;
	if (child == null)
	    return parentNew;
	childNew = copyTreeAtRoot(child);
	parentNew.firstchild = childNew;
	childNew.parent = parentNew; 

	prev     = child;
	child    = child.nextsibling;
	prevNew  = childNew;
	while (child != null) {
	    childNew = copyTreeAtRoot(child);
	    childNew.parent = parentNew; 
	    prevNew.nextsibling = childNew;

	    prev  = child;
	    child = child.nextsibling;
	    prevNew  = childNew;
	    childNew = childNew.nextsibling;
	}
	return parentNew;
    }

    public String toStringShort(){
	StringBuffer str = new StringBuffer();
	treeToString(getRootNode(), str);
	return str.toString();
    }

    public String toString() {
	Pattern pattern = this;
	StringBuffer str = new StringBuffer();
	str.append("<pattern"+
		   " id="+pattern.getId()+
		   " size="+pattern.getSize()+
		   " rmblen="+pattern.getRmbLength()+
		   " rootocc="+pattern.getRootOcc()+
		   " rmocc="+pattern.getRmOcc()+
		   ">");
	if (PRINT_STYLE_FOLD)
	    str.append("\n  ");
	treeToString(getRootNode(), str);
	str.append("</pattern>");
	return str.toString();
    }

    void treeToString(Snode node, StringBuffer str){

        str.append(node.toNodeNameStr());

        Snode nextnode = node.getFirstChild();
	boolean isFirstChild = true;
	boolean hasChild = (nextnode != null);

	if (hasChild)
	    str.append("[");
        while (nextnode != null){
	    if (!isFirstChild) str.append(", ");
            treeToString(nextnode, str);
            nextnode = nextnode.getNextSibling();
	    isFirstChild = false;
        }
	if (hasChild)
	    str.append("]");
	String value = node.getNodeValue();
	if (value != null) {
	    str.append("[\""+PrintDomTree.normalizeText(value)+"\"]");
	}
    }

    public static void main(String[] args){
        try{
	    int argc = args.length;
	    int numPat=0;
	    System.out.println("Pattern: Testing");
	    Pattern pat0 = new Pattern("root");

	    System.out.println(0+" => "+pat0.toString());
	    for (int i = 0; i < args.length; ) {
		int depth = Integer.parseInt(args[i++]);
		if (i >= argc)
		    Util.errorExit("depth="+depth+": A label required!");
		String label = new String(args[i++]);
		System.out.println("\nPattern: Expand with "+
				   "<"+depth+", "+label+">");
		pat0.expand(depth, new Snode(label));
		System.out.println(numPat+" => "+pat0.toString());
		numPat++;
	    }

	}
        catch(Exception e){
            System.err.println(e);
            System.exit(1);
        }
    }
}
