//
// Util.java
// $Id: Util.java,v 1.1.1.1 2004/03/24 12:19:00 arim Exp $
//

/*
 * Util.java
 * - Created by Hiroki Arimura on 10 DEC 01
*/

import java.io.*;
import java.lang.*;
import java.util.*;
import java.math.*;
import java.text.*;

public class Util {

  /************************************************************
   * Global Switches
   ************************************************************/
  public static boolean VERSION_MID        = true; 
  public static boolean ENABLE_DEBUG_PRINT = true;
  public static boolean ENABLE_XML_FORMAT  = false; 

  static long totaltime = 0;	// h^nԂvZ
  static long startTime = 0;	//Warning: Not to be nested
  
  /**********************************************************************
   * Utility routines
   **********************************************************************/

  public static void print(String str) {
    if (ENABLE_DEBUG_PRINT) {
      System.out.print(str);
      System.out.flush();
    }
  }

  public static void print() {
    if (ENABLE_DEBUG_PRINT) {
      System.out.flush();
    }
  }

  public static void println(String str) {
    if (ENABLE_DEBUG_PRINT) {
      System.out.println(str);
      System.out.flush();
    }
  }

  public static void println(Exception e) {
    if (ENABLE_DEBUG_PRINT) {
      System.out.println(e);
      System.out.flush();
    }
  }

  public static void println() {
    if (ENABLE_DEBUG_PRINT) {
      System.out.println();
    }
  }

  public static void errorExit(String str) {
    System.out.println("@error: " + str);
    System.exit(0);
  }

  public static void exit() {
    System.exit(0);
  }

    public static void task(String str) {
	System.out.println("");
	System.out.println("" + str);
    }

  public static void print_nchar(int n, char c)
    {
      for (int i = 0; i < n; i++)
	System.out.print(c);
    }

  static int indentWidth = 4;
  public static void setIndentWidth(int w) {
    indentWidth = w;
  }

  public static String indentString(int depth) {
      return ncharString(indentWidth*depth, ' ');
  }

  public static String ncharString(int n, char c) {
      StringBuffer buf = new StringBuffer();
      for (int i = 0; i < n; i++)
	  buf.append(c);
      return buf.toString();
  }

/*
 * Number Type conversion
 */

  public static String ratio_form(double x, int ncol) {
    String fx;
    NumberFormat nf = NumberFormat.getNumberInstance();
    nf.setMaximumFractionDigits(ncol);
    fx = nf.format(x);
    return fx;
  }

  public static String int_form(int x, int ncol) {
    String fx;
    NumberFormat nf = NumberFormat.getNumberInstance();
    nf.setMinimumIntegerDigits(ncol);
    nf.setMaximumIntegerDigits(ncol);
    fx = nf.format(x);
    return fx;
  }

  public static String boolean_form(boolean x, int ncol) {
    String fx;
    int valx; 
    if (x)
	valx = 1;
    else
	valx = 0;
    NumberFormat nf = NumberFormat.getNumberInstance();
    nf.setMinimumIntegerDigits(ncol);
    nf.setMaximumIntegerDigits(ncol);
    fx = nf.format(valx);
    return fx;
  }

  public static double a2double(String str) {
    return (new Double(str)).doubleValue();
  }

  public static String double2a(double num) {
    return (new Double(num)).toString();
  }
  public static int a2int(String str) {
    return (new Integer(str)).intValue();
  }

  public static String int2a(int num) {
    return (new Integer(num)).toString();
  }

}//End of Debug class

/** Cputime: Timing
 */
class CpuTimer {
    //Constants
    static final long UNDEF    = -1;
    static final boolean ON = true;
    static final boolean OFF= false;
    
    //Members 
    long gStartTime = UNDEF;
    long gElapsedTime = UNDEF;
    boolean isWorking = false;

    CpuTimer() {
	init();
    }
    
    CpuTimer(boolean StartFlag) {
	//Usage: timer = new CpuTimer(CpuTimer.CPU_ON);
	init();
	if (StartFlag)
	    start();
    }
    
    public void init() {
	gStartTime   = UNDEF;
	gElapsedTime = 0;
	isWorking    = false;
    }

    public long start() {
	gStartTime   = _currentTime();
	isWorking    = true;
	//Note: gElapsedTime is not changed
	return gStartTime; 
    }

    public long stop() {
	gElapsedTime = gElapsedTime + (_currentTime() - gStartTime);
	isWorking    = false;
	return gElapsedTime;
    }

    public void update() {
	if (isWorking) {
	    long curtime = _currentTime();
	    gElapsedTime = gElapsedTime + (curtime - gStartTime);
	    gStartTime   = curtime;
	    //Note: Timer is still working.
	}
    }

    public long getElapsedTime() {
	update();
	//Note: Timer is still working.
	return gElapsedTime;
    }

    public boolean isWorking() {
	return isWorking;
    }

    public void print() {
	update();
	System.out.println(
          "@time \t"+
	  Util.ratio_form((double)gElapsedTime * 0.001, 3)+" (sec)");
	//Note: Timer is still working.
    }

    public void print(String prefix) {
	update();
	System.out.println(
          prefix+
	  Util.ratio_form((double)gElapsedTime * 0.001, 3)+" (sec)");
	//Note: Timer is still working.
    }

    //static

    private static long _currentTime() {
	Date date = new Date();
	return date.getTime();
    }

}//CpuTimer


class SymbolTable{
  private HashMap nameTable;	// ^O̕\
  private HashMap countTable;	// ^OƂ̏o̕\
  private int count = 0;	// nbV\̃JE^

  public SymbolTable() 
  {
    nameTable  = new HashMap();
    countTable = new HashMap();
    count = 0;
  }

  public void clear() {
    nameTable.clear();
    countTable.clear();
    count = 0;
  }

  //  name ɈӂȎʔԍ^郁\bh
  // - ɓo^Ă΁CʔԍԂ
  // - ܂o^ĂȂΓo^C炽ȎʔԍԂD
  public int register(String name){

    Integer id;

    if (nameTable.containsKey(name) == true){
      // name ȑOɏoĂD
      id = (Integer) nameTable.get(name);
      if (id == null) {
	System.out.println("register: error id==null");
	System.exit(0);
      }
      //oXV
      Integer co = (Integer)countTable.get(name);
      if (co == null) {
	System.out.println("register: error co==null");
	System.exit(0);
      }
      //Replace the old value with the new one.
      countTable.put(name, (new Integer(co.intValue() + 1)));
    }
    else {
      //name ߂ďoD
      //Oo^
      id = new Integer(count);
      nameTable.put(name, id);
      //oJE^o^
      Integer co = new Integer(1); // The number 1 means the first occ.
      countTable.put(name, co);
      //IdXV
      count++;
    }

    return(id.intValue());
  }

  // name o^Ă邩ǂ^UlԂ
  public boolean contains(String name)
  {
    return nameTable.containsKey(name); 
  }

  // ^ꂽnode ̎ʔԍԂ\bh
  // o^ĂȂ΁C̒l(-1)ԂD
  public int lookup(String name){

    int id;

    if (nameTable.containsKey(name) == true){
      id = ((Integer)nameTable.get(name)).intValue();
    }
    else {
      id = (-1); 
    }
    return id;
  }

  public void reportCount(){
    Set entries = countTable.entrySet();
    Iterator iter = entries.iterator();
    while (iter.hasNext()) {
      Map.Entry entry = (Map.Entry) iter.next();
      String key    = (String) entry.getKey();
      Integer value = (Integer) entry.getValue();
      System.out.println("\t" + key + "\t" + value);
    }
  }

  public int getSize()
  {
    return nameTable.size();
  }
}

